#
/*
 *		M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *				    D P R O C S . C
 *
 *				Douglas P. Kingston III
 *				  Michael John Muuss
 *
 *				Part of the MDQS Daemon
 *
 *				    Revision History
 *				    ================
 *	13 Sep 82  DPK	Added code to support Waveoffs.
 *
 *	14 Sep 82  DPK	Changed handling of unopenable devices.
 *
 *	15 Sep 82  DPK	Added argument fiddling for devstart()
 *
 *	22 Sep 82  DPK	Changed exec code to allow invoking daemons with args
 *
 *	17 Nov 82  DPK	Added code to cause filecontrol to awaken daemon upon
 *			death of child.  Added request name to pidfile.
 *			Added termall() shutdown code.
 *
 *	30 Mar 83  DPK  Added fixes and dv_failed handling as suggested
 *			by Dave Wilson at Boeing.  Also other minor changes
 *			to handling of server specification.
 */

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include "queue.h"
#include "qreplys.h"
#include <pwd.h>
#ifdef BSD42
#include <sys/wait.h>
#endif

extern	struct device *devhead;
extern	int Procwait;
extern	int Openwait;
extern	int Delaytime;
extern	int Waveoff;
extern	int Maxfailures;
extern	char **origargv;

extern	FILE	*msgopen();
extern	char	*xmalloc();
extern	int	termchild();

static int	childpid;
int	restartchild();
int	flushchild();

/*
 *	D E V S T A R T ( )
 *
 *	Activate an idle device.
 */
devstart (dp)
register struct device *dp;
{
	register struct request *rp;
	int	pip[2];		/* for pipe() system call return */
	int	fd;

	if( dp->dv_pid != 0 )		/* Paranoid. */
		complain( "devstart called on busy device (%s, %d)",
			  dp->dv_dname, dp->dv_pid );

	else if( dp->dv_actreq == NULL )	/* Paranoid again. */
		complain( "devstart called with no request" );

	else if (( dp->dv_pid = fork ()) == 0 ) {
		int	padlen;
		struct passwd *pw;

		/*
		 *  CHILD - attempt to open device, fork server
		 */

		/* Change up command line for benefit of ps(1). */
		padlen = strlen (origargv[0]) - strlen (dp->dv_dname);
		if (padlen > 0)
			sprintf (origargv[0], "%s%-*.*s", dp->dv_dname,
				padlen, padlen, " control");
		else
			sprintf (origargv[0], "%.*s",
				 strlen (origargv[0]), dp->dv_dname);

		rp = dp->dv_actreq;
		childpid = 0;
#ifdef BSD42
		if ((pw = getpwuid(rp->r_uid)) == 0) {
			char *cp = "XUser not in /etc/passwd";

			cp[0] = RP_FATAL;
			inform(rp, dp, cp);
			_exit(RP_FATAL);
		}			
#endif BSD42

		/*
		 *  Setup for restart and flush signalling
		 */
		signal (SIGEMT, restartchild);
		signal (SIGTERM, flushchild);

		/*
		 *  Record the PID for restarts and flushes
		 */
		devactive (dp, getpid(), "Prenatal open");

		/*
		 *  Open control file on fd 0
		 */
		if ((fd = open (rp->r_name, READ)) < 0) {
			char *cp = "XUnable to access control file";

			complain ("Control file %s: %s", rp->r_name,
				   sys_errlist [errno]);
			cp[0] = RP_HOLD;
			inform (rp, dp, cp);
			devactive(dp, 0, "Idle");
			_exit (RP_HOLD);
		}
		if (fd != 0) {
			dup2 (fd, 0);
			close (fd);
		}

debug( "devstart() opening %s", dp->dv_device );
		/*
		 *  Try and get the device, if not twiddle our thumbs,
		 *  and return some time later, open will be retryed.
		 */
		if (( fd = open (dp->dv_device, RW)) < 0
		 && ( fd = open (dp->dv_device, WRITE)) < 0) {
debug( "devstart() failed to open %s", dp->dv_device );			
#ifdef BSD42
			if (initgroups (pw->pw_name, rp->r_gid) < 0) {
				char *cp = "XInitgroups failed";

				cp[0] = RP_FATAL;
				inform(rp, dp, cp);
				_exit(RP_FATAL);
			}
#else BSD42
			setgid( rp->r_gid );	/* Activates incoming sigs */
#endif BSD42
			setuid( rp->r_uid );
			sleep( Openwait );	/* To prevent "thrashing" */
		 	devactive(dp, 0, "Idle");
			_exit(RP_NOPEN);
		}
		if (fd != 1) {
			dup2 (fd, 1);
			close (fd);
		}

		/*
		 *  Start device service program
		 */
		if ( pipe (pip) < 0 ) {
			complain ("devstart: pipe failed: %s",
				   sys_errlist[errno]);
			devactive(dp, 0, "Idle");
			_exit (RP_RESTART);
		}
		childpid = fork();
		if ( childpid < 0 )  {
			complain ("devstart: fork failed: %s",
				   sys_errlist[errno]);
			devactive(dp, 0, "Idle");
			_exit (RP_RESTART);
		}

		/*
		 *  The child process sets up the WRITE end of the
		 *  pipe and then execs the server.  The parent
		 *  sets up the READ end and waits for at lease one
		 *  byte of data from the pipe, anything less indicates
		 *  a fatal error on the part of the child.  The data
		 *  returned is then used as the exitcode for the
		 *  parent.  The parent does NOT wait for the child.
		 *	fd 0 - Control file
		 *	fd 1 - Device
		 *	fd 2 - Error console (/dev/console usually)
		 *	fd 3 - PIPE to daemon's son
		 *	fd 4 - Device status file
		 *
		 *	All other filedescriptors closed.
		 */
		if ( childpid == 0 )  {
			char	resultcode;
			char	**argv;
			char	execbuf[PATHSIZ];
			char	*prog;

			/*
			 *  Child of Devstart  ( Grandchild of the Daemon )
			 */

#if V7 | SYS3
			setpgrp ();
#endif
#if BSD42 | BSD41 | BSD41c
			setpgrp (0, getpid());
#endif
#if BRLUNIX
			setpg (0, getpid());
#endif
			signal( SIGTERM, SIG_DFL );
			signal( SIGEMT, SIG_DFL );
#ifdef BSD42
			if (initgroups (pw->pw_name, rp->r_gid) < 0) {
				char *cp = "XInitgroups failed";

				cp[0] = RP_FATAL;
				inform(rp, dp, cp);
				_exit(RP_FATAL);
			}
#else BSD42
			setgid( rp->r_gid );
#endif BSD42
			setuid( rp->r_uid );
			close (pip[READ]);
			dup2 (pip[WRITE], 3);
			openstat(dp, 4);	/* Open status file on 4 */

#ifndef _NFILE
			for( fd = getdtablesize()-1; fd > 4; fd-- )
#else
			for( fd = _NFILE-1; fd > 4; fd-- )
#endif
				close( fd );

			strncpy( execbuf, dp->dv_server, PATHSIZ-1 );
			if (parse( execbuf, &argv ) < 0) {
				complain ("Bad server specification on %s (MAXFIELDS exceeded)",
					  dp->dv_dname);
				goto bombout;
			}

			/*
			 *  For the benefit of PS...
			 *  8 is the size of " server\0", see strcat below.
			 */
			prog = argv[0];
			argv[0] = xmalloc (strlen (dp->dv_dname) + 8);
			strcpy (argv[0], dp->dv_dname);
			strcat (argv[0], " server");

			execv( prog, argv );
			complain ("devstart exec error, %s: %s",
				 prog, sys_errlist[errno]);
bombout:
			resultcode = RP_NEXEC;
			write( 3, &resultcode, 1 );
			_exit( -1 );
		} else {
			char	resultstring[BUFSIZ];
			int	nread;

			/*
			 *	Parent
			 */
			close( pip[WRITE] );
			dup2( 2, 1 );		/* Closes device */

#ifdef BSD42
			if (initgroups (pw->pw_name, rp->r_gid) < 0) {
				char *cp = "XInitgroups failed";

				cp[0] = RP_FATAL;
				inform(rp, dp, cp);
				_exit(RP_FATAL);
			}
#else BSD42
			setgid( rp->r_gid );	/* Activates incoming sigs */
#endif BSD42
			setuid( rp->r_uid );

			while ((nread = read (pip[READ], resultstring,
				  (sizeof resultstring) - 1)) < 1) {
				if( nread < 0 && errno == EINTR )
				  	continue;
				complain ("devstart: bad read (%d, %d, %s)",
					   nread, errno, dp->dv_server);
				devactive(dp, 0, "Idle");
				strcpy (resultstring, "XBad read from server");
				resultstring[0] = RP_DIED;
				inform (rp, dp, resultstring);
				_exit (RP_DIED);
			}
			resultstring[nread] = 0;
			devactive(dp, 0, "Idle");
			signal( SIGEMT, SIG_IGN );
			signal( SIGTERM, SIG_IGN );
debug("Devstart exits with 0%o, '%s'", resultstring[0], &resultstring[1]);
			inform (rp, dp, resultstring);
			_exit (resultstring[0]);
		}
	}
	/*
	 *	If the fork failed, we need to reset the pid field
	 *	to cause this request to be restarted.
	 */
	if( dp->dv_pid == -1 )
		dp->dv_pid = 0;
}

/*
 *			P W A I T
 *
 *	Collects the devstart() process and cleans up after
 *	a request.  Will restart a request if necessary.
 */
pwait ()
{
	register struct device *dp;
	register int	pid;
	register struct request *rp;
	struct	status status;

#ifndef BSD42
	int	retstat;

	alarm (Procwait);
	while ((pid = wait (&retstat)) != -1) {
		alarm (0);
		status.ret = (retstat>>8) & 0xff;
		status.pstat = retstat & 0xff;
#else
	union wait uwait;

	while ((pid = wait3(&uwait, WNOHANG, (struct rusage *)0)) > 0) {
		status.ret = uwait.w_retcode;
		status.pstat = uwait.w_coredump;
#endif
debug ("pwait: just swallowed %d (%o)", pid, status);

		/* Find device this pid was associated with */
		for (dp = devhead; dp != NULL; dp = dp->dv_nextdev)
			if (dp->dv_pid == pid)
				break;
		if (dp->dv_pid != pid) {
			/* PID was not found in any of device structure */
debug ("Got unknown child %d", pid);
#ifndef BSD42
			alarm( Procwait );
#endif
			continue;
		}

debug( "pwait(): clear %d from %s (ret was %d/%d)", dp->dv_pid,
	dp->dv_dname, status.ret, status.pstat );

		dp->dv_pid = 0;
		rp = dp->dv_actreq;
		/*
		 *  Assumption:  Daemon is synchronous and
		 *  therefore this job will not get restarted
		 *  until this code completes (daemon is single
		 *  thread).  If daemon were multi-thread, we
		 *  would have to move this next line to the end.
		 */
		rp->r_status &= ~SRUN;

		if (status.pstat != 0) {
			char msg[LINESIZ];

			/* Horrible death */
			complain ("pwait: horrible death of %s, status %d",
				   dp->dv_server, status.pstat);
			complain ("pwait: job %s held, device %s disabled",
				rp->r_name, dp->dv_dname);
			sprintf( msg,
				"%cThe server daemon died from a signal %d",
				RP_DIED, status.pstat );
			inform( rp, dp, msg );
			hold( rp->r_name );
			delfromq(dp->dv_actque, rp);
			zapreq( rp );
			free( rp );
			dp->dv_status |= D_FAILED;	/* Take it down */
			devactive (dp, 0, "Idle");	/* Dead child may not */
		} else {
			time_t timenow;

			delfromq(dp->dv_actque, rp); /* Will requeue or zap */

			/* Good status, look at return code ... */
			switch (status.ret) {
			case RP_HOLD:
				/* Request causes sever problem, hold it */
				hold( rp->r_name );
				/*** FALL THROUGH ***/

			case RP_FLUSH:
				/* Request flushed */
debug( "Flushing %s", rp->r_name );
				/* FALL THROUGH */

			case RP_OK:
				/* Everything's hunky-dorey */
				zapreq( rp );
				unlink( rp->r_name );
				free( rp );
				dp->dv_failed = 0;
				break;

			case RP_NOPEN:
	 			/* Couldn't open device, scandevs will retry */
			case RP_NEXEC:
				/* Couldn't exec server, scandevs will retry */
				dp->dv_failed++;

				/*
				 *  Don't use this device if it has
				 *  failed too many times, disable for now.
				 *  Currently this can only be cleared by
				 *  removing the device and then recreating.
				 */
				if ( Maxfailures
				   && dp->dv_failed >= Maxfailures )
				{
					complain( "device %s disabled, (failed %d times)",
						dp->dv_dname, dp->dv_failed );
					dp->dv_status |= D_FAILED;
				}
				/** Fall Through! **/

			case RP_RESTART:
				/* Some error, restart same request */
				if( rp->r_status & SKILL ) {
					char *flushmsg = "XFlushing (SKILL)";

					/* Catch SKILL bits */
debug( "Flushing %s", rp->r_name );
					flushmsg[0] = RP_KILLED;
					inform( rp, dp, flushmsg );
					zapreq( rp );
					unlink( rp->r_name );
					free( rp );
					break;
				}
debug( "Restart %s", rp->r_name );
				if( queue( rp ) < 0 ) {
					char	*lostmsg = 
			"XIt was unable to be requeued after a restart.\n";

					lostmsg[0] = RP_FATAL;
					complain(
			"restart of %s failed -- no queue", rp->r_name );
					inform( rp, dp, lostmsg );
					zapreq( rp );
					hold( rp->r_name );
					free( rp );
				}
				dp->dv_failed = 0;
				break;

			case RP_DELAY:
				/* Net access fail, delay retry */
				if( rp->r_status & SKILL ) {
					char *flushmsg = "XFlushing (SKILL)";

					/* Catch SKILL bits */
debug( "Flushing %s", rp->r_name );
					flushmsg[0] = RP_KILLED;
					inform( rp, dp, flushmsg );
					zapreq( rp );
					unlink( rp->r_name );
					free( rp );
					break;
				}
debug( "Netfail %s", rp->r_name );
				time(&timenow);
				rp->r_tdelay = timenow + (long)Delaytime;
				rp->r_status |= SDELAY;
				rupdate(rp);
				if( queue( rp ) < 0 ) {
					char	*lostmsg = 
			"XIt was unable to be requeued after a netfail.\n";

					lostmsg[0] = RP_FATAL;
					complain(
			"requeue of %s failed -- no queue", rp->r_name );
					inform( rp, dp, lostmsg );
					zapreq( rp );
					hold( rp->r_name );
					free( rp );
				}
				dp->dv_failed = 0;
				break;

			default:
				/* Shouldn't happen... */
				complain( "bad completion code 0%o: dev %s, server %s",
					  status.ret, dp->dv_dname,
					  dp->dv_server );
				/* FALL THROUGH */

			case RP_DIED:
			case RP_FATAL:
				/* Request failed, pitch it */
				zapreq( rp );
				unlink( rp->r_name );
				free( rp );
				dp->dv_failed++;
				break;
			}
		}

		/* Clear out the current device */
		dp->dv_server = NULL;
		dp->dv_actreq = NULL;
		dp->dv_actque = NULL;
		Waveoff = 0;

#ifndef BSD42
		alarm( Procwait );
#endif
	} /* end of while */
#ifndef BSD42
	alarm(0);
#endif
}

/*
 *	T E R M A L L ( )
 *
 *	Send RESTART signal to all children of the daemon.
 *	This will cause the children to stop and return to
 *	be restarted, but because we are about to shutdown,
 *	nothing is restarted at this time.
 */
termall()
{
	register struct device *dp;

	for( dp = devhead; dp != NULL; dp = dp->dv_nextdev ) {
debug("termall(%s, %d)", dp->dv_dname, dp->dv_pid);
		if( dp->dv_pid > 1 )
			kill( dp->dv_pid, SIGEMT );
	}
	sync();
}

/*
 *	RESTARTCHILD ( )
 *
 *	This function will be activated by either the qrestart
 *	program or by the MDQS daemon for a MDQS shutdown request.
 */
restartchild()
{
	signal( SIGEMT, restartchild );
debug("restarting %d from %d", childpid, getpid());
	if( childpid > 1 )		/* Paranoid */
#if V6 | DPWUNIX
		kill( childpid, SIGEMT );
#endif
#if BSD42 | BSD41 | BSD41c | BRLUNIX
		killpg( childpid, SIGEMT );
#endif
#if SYS3 | V7
		kill( -childpid, SIGEMT );
#endif
}

/*
 *	FLUSHCHILD ( )
 *
 *	This function will be activated the qflush program.
 */
flushchild()
{
	signal( SIGTERM, flushchild );
debug("terminating %d from %d", childpid, getpid());
	if( childpid > 1 )		/* Paranoid */
#if V6 | DPWUNIX
		kill( childpid, SIGTERM );
#endif
#if BSD42 | BSD41 | BSD41c | BRLUNIX
		killpg( childpid, SIGTERM );
#endif
#if SYS3 | V7
		kill( -childpid, SIGTERM );
#endif
}

/*
 *	I N F O R M
 */
inform (rp, dp, result)
register struct request *rp;
struct device *dp;
register char *result;
{
	register FILE	*mf;

debug("inform on %s", rp->r_name);
	log ("%s on %s: %s (%03o)", rp->r_name, dp->dv_dname, &result[1],
		result[0]&0377);
	switch (result[0]) {
	case RP_RESTART:	/* Don't inform on restart errors */
	case RP_DELAY:
	case RP_NEXEC:
	case RP_NOPEN:
		return;

	case RP_FLUSH:
		rp->r_options |= O_MAIL;
		if(( mf = msgopen( rp )) == NULL )
			return;
		fprintf( mf, "Your %s request was flushed from device %s.\n",
			 dp->dv_actque->q_name, dp->dv_dname );
		if (result[1]) {
			fputs (&result[1], mf);
			fputc ('\n', mf);
		}
		break;

	case RP_OK:
		if( (dp->dv_status & D_SKIPMSG)
		  || (mf = msgopen( rp )) == NULL )
			return;
		fprintf( mf, "Your %s request completed successfully.\n",
			 dp->dv_actque->q_name );
		if (result[1]) {
			fputs (&result[1], mf);
			fputc ('\n', mf);
		}
		break;

	case RP_KILLED:
		rp->r_options |= O_MAIL;
		if(( mf = msgopen( rp )) == NULL )
			return;
		fprintf( mf,
		    "Your %s request was cancelled from device %s.\n",
			dp->dv_actque->q_name, dp->dv_dname );
		if (result[1]) {
			fputs (&result[1], mf);
			fputc ('\n', mf);
		}
		break;

	case RP_DIED:
	case RP_HOLD:
	case RP_FATAL:
		rp->r_options |= O_MAIL;
		if(( mf = msgopen( rp )) == NULL )
			return;
		fprintf( mf, "Your %s request failed on device %s.\n",
			dp->dv_actque->q_name, dp->dv_dname );
	   	if (result[1]) {
	   		fputs (&result[1], mf);
			fputc ('\n', mf);
		}
		fprintf( mf,
	  "You may try resubmitting your request, but if it fails again\n" );
		fprintf( mf,
	  "please consult a member of the support staff for help\n" );
		break;

	default:
		complain( "bad result code 0%o: dev %s, server %s",
			  result[0], dp->dv_dname, dp->dv_server );
		rp->r_options |= O_MAIL;
		if(( mf = msgopen( rp )) == NULL )
			return;
		fprintf( mf, "Your %s request on device %s finished\n",
			dp->dv_actque->q_name, dp->dv_dname );
		fprintf( mf, "with an unknown completion completion code.\n");
	   	if (result[1]) {
	   		fputs (&result[1], mf);
			fputc ('\n', mf);
		}
		fprintf( mf,
	    "Please consult a member of the support staff for help\n" );
		break;
	}
	fclose (mf);
}

delfromq(qp, therp)
struct queue *qp;
register struct request *therp;
{
	register struct request *rp;
	register struct request *prevrp;

	rp = qp->q_readyreq;
	if (rp == NULL)
		goto not_here;

	if (rp == therp) {
		/* We found it at the head, easy case! */
		qp->q_readyreq = rp->r_nextreq;
		qp->q_status |= Q_UPDATE;
		return(1);
	}
	for (prevrp = rp, rp = rp->r_nextreq; rp != NULL;
	     prevrp = rp, rp = rp->r_nextreq)
		if (rp == therp) {
			/* Delete from the middle */
			prevrp->r_nextreq = rp->r_nextreq;
			qp->q_status |= Q_UPDATE;
			return (1);
		}
not_here:
	complain ("pwait: request %s no on queue %s",
		therp->r_name, qp->q_name);
	return (0);
}
